import abc
import asyncio
import logging
import time

from ydb import credentials
from ydb import issues

logger = logging.getLogger(__name__)
YDB_AUTH_TICKET_HEADER = "x-ydb-auth-ticket"


class AtMostOneExecution(object):
    def __init__(self):
        self._can_schedule = True
        self._lock = asyncio.Lock()

    async def wrapped_execution(self, callback):
        async with self._lock:
            try:
                await callback()
            except Exception:
                pass
            finally:
                self._can_schedule = True

    def submit(self, callback):
        if self._can_schedule:
            self._can_schedule = False
            asyncio.create_task(self.wrapped_execution(callback))


class AbstractExpiringTokenCredentials(credentials.AbstractExpiringTokenCredentials):
    def __init__(self):
        super(AbstractExpiringTokenCredentials, self).__init__()
        self._token_lock = asyncio.Lock()
        self._tp = AtMostOneExecution()

    @abc.abstractmethod
    async def _make_token_request(self):
        pass

    async def get_auth_token(self) -> str:
        for header, token in await self.auth_metadata():
            if header == YDB_AUTH_TICKET_HEADER:
                return token
        return ""

    async def _refresh_token(self, should_raise=False):
        current_time = time.time()

        try:
            self.logger.debug(
                "Refreshing token async, current_time: %s, expires_in: %s", current_time, self._expires_in
            )

            token_response = await self._make_token_request()
            self._update_token_info(token_response, current_time)

            self.logger.info("Token refreshed successfully async, expires_in: %s", self._expires_in)
            self.last_error = None

        except Exception as e:
            self.last_error = str(e)
            self.logger.error("Failed to refresh token async: %s", e)
            if should_raise:
                raise issues.ConnectionError(
                    "%s: %s.\n%s" % (self.__class__.__name__, self.last_error, self.extra_error_message)
                )

    async def token(self):
        if self._is_token_valid():
            if self._should_refresh():
                self._tp.submit(self._refresh_token)

            return self._cached_token

        async with self._token_lock:
            if self._is_token_valid():
                return self._cached_token

            await self._refresh_token(should_raise=True)

        return self._cached_token

    async def auth_metadata(self):
        return [(credentials.YDB_AUTH_TICKET_HEADER, await self.token())]



# -*- coding: utf-8 -*-
from .credentials import AbstractExpiringTokenCredentials
from ydb.oauth2_token_exchange.token_source import TokenSource
from ydb.oauth2_token_exchange.token_exchange import Oauth2TokenExchangeCredentialsBase
import typing

try:
    import aiohttp
except ImportError:
    aiohttp = None


class Oauth2TokenExchangeCredentials(AbstractExpiringTokenCredentials, Oauth2TokenExchangeCredentialsBase):
    def __init__(
        self,
        token_endpoint: str,
        subject_token_source: typing.Optional[TokenSource] = None,
        actor_token_source: typing.Optional[TokenSource] = None,
        audience: typing.Union[typing.List[str], str, None] = None,
        scope: typing.Union[typing.List[str], str, None] = None,
        resource: typing.Optional[str] = None,
        grant_type: str = "urn:ietf:params:oauth:grant-type:token-exchange",
        requested_token_type: str = "urn:ietf:params:oauth:token-type:access_token",
    ):
        assert aiohttp is not None, "Install aiohttp library to use OAuth 2.0 token exchange credentials provider"
        super(Oauth2TokenExchangeCredentials, self).__init__()
        Oauth2TokenExchangeCredentialsBase.__init__(
            self,
            token_endpoint,
            subject_token_source,
            actor_token_source,
            audience,
            scope,
            resource,
            grant_type,
            requested_token_type,
        )

    async def _make_token_request(self):
        params = self._make_token_request_params()
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        timeout = aiohttp.ClientTimeout(total=2)
        async with aiohttp.ClientSession(timeout=timeout) as session:
            async with session.post(self._token_endpoint, data=params, headers=headers) as response:
                self._process_response_status_code(await response.text(), response.status)
                return self._process_response_json(await response.json())



from __future__ import absolute_import, unicode_literals

import collections
import datetime
import itertools
import logging

import ydb
from .errors import DatabaseError


LOGGER = logging.getLogger(__name__)


STR_QUOTE_MAP = (
    ("\\", "\\\\"),
    ("'", r"\'"),
    ("\0", r"\x00"),
    # To re-check: \b \f \r \n \t
)


def render_str(value):
    for r_from, r_to in STR_QUOTE_MAP:
        value = value.replace(r_from, r_to)
    return "'" + value + "'"


def render_date(value):
    return "Date({})".format(render_str(value.isoformat()))


def render_datetime(value):
    # TODO: is there a better solution for this?
    return "DateTime::MakeDatetime(DateTime::ParseIso8601({}))".format(render_str(value.isoformat()))


def render(value):
    if value is None:
        return "NULL"
    if isinstance(value, str):
        return render_str(value)
    if isinstance(value, datetime.datetime):
        return render_datetime(value)
    if isinstance(value, datetime.date):
        return render_date(value)
    return repr(value)


def render_sql(sql, parameters):
    if not parameters:
        return sql

    assert sql.count("?") == len(parameters), "num of placeholders != num of params"

    quoted_params = [render(param) for param in parameters]
    quoted_params += [""]
    sql_pieces = sql.split("?")
    assert len(sql_pieces) == len(quoted_params)
    return "".join(piece for pair in zip(sql_pieces, quoted_params) for piece in pair if piece)


def named_result_for(column_names):
    # TODO fix: this doesn't allow columns names starting with underscore, e.g. `select 1 as _a`.
    return collections.namedtuple("NamedResult", column_names)


def _get_column_type(type_obj):
    return str(type_obj)


def get_column_type(type_obj):
    return _get_column_type(ydb.convert.type_to_native(type_obj))


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self.description = []
        self.arraysize = 1
        self.logger = LOGGER
        self.rows = None
        self._rows_prefetched = None

    def execute(self, sql, parameters=None):
        fsql = render_sql(sql, parameters)
        self.logger.debug("execute sql: %s", fsql)
        try:
            chunks = self.connection.driver.table_client.scan_query(fsql)
        except ydb.Error as e:
            raise DatabaseError(e.message, e.issues, e.status)

        self.description = []

        rows = self._rows_iterable(chunks)
        # Prefetch the description:
        try:
            first_row = next(rows)
        except StopIteration:
            pass
        else:
            rows = itertools.chain((first_row,), rows)
        if self.rows is not None:
            rows = itertools.chain(self.rows, rows)

        self.rows = rows

    def _rows_iterable(self, chunks_iterable):
        description = None
        try:
            for chunk in chunks_iterable:
                if description is None and len(chunk.result_set.rows) > 0:
                    description = [
                        (
                            col.name,
                            get_column_type(col.type),
                            None,
                            None,
                            None,
                            None,
                            None,
                        )
                        for col in chunk.result_set.columns
                    ]
                    self.description = description
                for row in chunk.result_set.rows:
                    # returns tuple to be compatible with SqlAlchemy and because
                    #  of this PEP to return a sequence: https://www.python.org/dev/peps/pep-0249/#fetchmany
                    yield row[::]
        except ydb.Error as e:
            raise DatabaseError(e.message, e.issues, e.status)

    def _ensure_prefetched(self):
        if self.rows is not None and self._rows_prefetched is None:
            self._rows_prefetched = list(self.rows)
            self.rows = iter(self._rows_prefetched)
        return self._rows_prefetched

    def executemany(self, sql, seq_of_parameters):
        for parameters in seq_of_parameters:
            self.execute(sql, parameters)

    def executescript(self, script):
        return self.execute(script)

    def fetchone(self):
        if self.rows is None:
            return None
        try:
            return next(self.rows)
        except StopIteration:
            return None

    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize

        return list(itertools.islice(self.rows, size))

    def fetchall(self):
        return list(self.rows)

    def nextset(self):
        self.fetchall()

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, column=None):
        pass

    def close(self):
        self.rows = None
        self._rows_prefetched = None

    @property
    def rowcount(self):
        return len(self._ensure_prefetched())



from google.protobuf import any_pb2 as _any_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_common_pb2 as _ydb_common_pb2
from protos import ydb_issue_message_pb2 as _ydb_issue_message_pb2
from protos import ydb_status_codes_pb2 as _ydb_status_codes_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CancelOperationRequest(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class CancelOperationResponse(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class ForgetOperationRequest(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class ForgetOperationResponse(_message.Message):
    __slots__ = ["issues", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ...) -> None: ...

class GetOperationRequest(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetOperationResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: Operation
    def __init__(self, operation: _Optional[_Union[Operation, _Mapping]] = ...) -> None: ...

class ListOperationsRequest(_message.Message):
    __slots__ = ["kind", "page_size", "page_token"]
    KIND_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    kind: str
    page_size: int
    page_token: str
    def __init__(self, kind: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ...) -> None: ...

class ListOperationsResponse(_message.Message):
    __slots__ = ["issues", "next_page_token", "operations", "status"]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    OPERATIONS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    next_page_token: str
    operations: _containers.RepeatedCompositeFieldContainer[Operation]
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., operations: _Optional[_Iterable[_Union[Operation, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...

class Operation(_message.Message):
    __slots__ = ["cost_info", "id", "issues", "metadata", "ready", "result", "status"]
    COST_INFO_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    ISSUES_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    READY_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    cost_info: _ydb_common_pb2.CostInfo
    id: str
    issues: _containers.RepeatedCompositeFieldContainer[_ydb_issue_message_pb2.IssueMessage]
    metadata: _any_pb2.Any
    ready: bool
    result: _any_pb2.Any
    status: _ydb_status_codes_pb2.StatusIds.StatusCode
    def __init__(self, id: _Optional[str] = ..., ready: bool = ..., status: _Optional[_Union[_ydb_status_codes_pb2.StatusIds.StatusCode, str]] = ..., issues: _Optional[_Iterable[_Union[_ydb_issue_message_pb2.IssueMessage, _Mapping]]] = ..., result: _Optional[_Union[_any_pb2.Any, _Mapping]] = ..., metadata: _Optional[_Union[_any_pb2.Any, _Mapping]] = ..., cost_info: _Optional[_Union[_ydb_common_pb2.CostInfo, _Mapping]] = ...) -> None: ...

class OperationParams(_message.Message):
    __slots__ = ["cancel_after", "labels", "operation_mode", "operation_timeout", "report_cost_info"]
    class OperationMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class LabelsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ASYNC: OperationParams.OperationMode
    CANCEL_AFTER_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    OPERATION_MODE_FIELD_NUMBER: _ClassVar[int]
    OPERATION_MODE_UNSPECIFIED: OperationParams.OperationMode
    OPERATION_TIMEOUT_FIELD_NUMBER: _ClassVar[int]
    REPORT_COST_INFO_FIELD_NUMBER: _ClassVar[int]
    SYNC: OperationParams.OperationMode
    cancel_after: _duration_pb2.Duration
    labels: _containers.ScalarMap[str, str]
    operation_mode: OperationParams.OperationMode
    operation_timeout: _duration_pb2.Duration
    report_cost_info: _ydb_common_pb2.FeatureFlag.Status
    def __init__(self, operation_mode: _Optional[_Union[OperationParams.OperationMode, str]] = ..., operation_timeout: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., cancel_after: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., labels: _Optional[_Mapping[str, str]] = ..., report_cost_info: _Optional[_Union[_ydb_common_pb2.FeatureFlag.Status, str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_monitoring_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v4.protos import ydb_monitoring_pb2 as protos_dot_ydb__monitoring__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17ydb_monitoring_v1.proto\x12\x11Ydb.Monitoring.V1\x1a\x1bprotos/ydb_monitoring.proto2\xb7\x01\n\x11MonitoringService\x12P\n\tSelfCheck\x12 .Ydb.Monitoring.SelfCheckRequest\x1a!.Ydb.Monitoring.SelfCheckResponse\x12P\n\tNodeCheck\x12 .Ydb.Monitoring.NodeCheckRequest\x1a!.Ydb.Monitoring.NodeCheckResponseBY\n\x1ctech.ydb.proto.monitoring.v1Z9github.com/ydb-platform/ydb-go-genproto/Ydb_Monitoring_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_monitoring_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034tech.ydb.proto.monitoring.v1Z9github.com/ydb-platform/ydb-go-genproto/Ydb_Monitoring_V1'
  _MONITORINGSERVICE._serialized_start=76
  _MONITORINGSERVICE._serialized_end=259
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
import importlib.util
import threading
import codecs
from concurrent import futures
import functools
import hashlib
import collections
import urllib.parse
from . import ydb_version

try:
    from . import interceptor
except ImportError:
    interceptor = None


_grpcs_protocol = "grpcs://"
_grpc_protocol = "grpc://"


def wrap_result_in_future(result):
    f = futures.Future()
    f.set_result(result)
    return f


def wrap_exception_in_future(exc):
    f = futures.Future()
    f.set_exception(exc)
    return f


def future():
    return futures.Future()


def x_ydb_sdk_build_info_header():
    return ("x-ydb-sdk-build-info", "ydb-python-sdk/" + ydb_version.VERSION)


def is_secure_protocol(endpoint):
    return endpoint.startswith("grpcs://")


def wrap_endpoint(endpoint):
    if endpoint.startswith(_grpcs_protocol):
        return endpoint[len(_grpcs_protocol) :]
    if endpoint.startswith(_grpc_protocol):
        return endpoint[len(_grpc_protocol) :]
    return endpoint


def parse_connection_string(connection_string):
    cs = connection_string
    if not cs.startswith(_grpc_protocol) and not cs.startswith(_grpcs_protocol):
        # default is grpcs
        cs = _grpcs_protocol + cs

    p = urllib.parse.urlparse(connection_string)
    b = urllib.parse.parse_qs(p.query)
    database = b.get("database", [])
    assert len(database) > 0

    return p.scheme + "://" + p.netloc, database[0]


# Decorator that ensures no exceptions are leaked from decorated async call
def wrap_async_call_exceptions(f):
    @functools.wraps(f)
    def decorator(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            return wrap_exception_in_future(e)

    return decorator


def check_module_exists(path: str) -> bool:
    try:
        if importlib.util.find_spec(path):
            return True
    except ModuleNotFoundError:
        pass
    return False


def get_query_hash(yql_text):
    try:
        return hashlib.sha256(str(yql_text, "utf-8").encode("utf-8")).hexdigest()
    except TypeError:
        return hashlib.sha256(str(yql_text).encode("utf-8")).hexdigest()


class LRUCache(object):
    def __init__(self, capacity=1000):
        self.items = collections.OrderedDict()
        self.capacity = capacity

    def put(self, key, value):
        self.items[key] = value
        while len(self.items) > self.capacity:
            self.items.popitem(last=False)

    def get(self, key, _default):
        if key not in self.items:
            return _default
        value = self.items.pop(key)
        self.items[key] = value
        return value

    def erase(self, key):
        self.items.pop(key)


def from_bytes(val):
    """
    Translates value into valid utf8 string
    :param val: A value to translate
    :return: A valid utf8 string
    """
    try:
        return codecs.decode(val, "utf8")
    except (UnicodeEncodeError, TypeError):
        return val


class AsyncResponseIterator(object):
    def __init__(self, it, wrapper):
        self.it = it
        self.wrapper = wrapper

    def cancel(self):
        self.it.cancel()
        return self

    def __iter__(self):
        return self

    def _next(self):
        return interceptor.operate_async_stream_call(self.it, self.wrapper)

    def next(self):
        return self._next()

    def __next__(self):
        return self._next()


class SyncResponseIterator(object):
    def __init__(self, it, wrapper):
        self.it = it
        self.wrapper = wrapper

    def cancel(self):
        self.it.cancel()
        return self

    def __iter__(self):
        return self

    def _next(self):
        res = self.wrapper(next(self.it))
        if res is not None:
            return res
        return self._next()

    def next(self):
        return self._next()

    def __next__(self):
        return self._next()


class AtomicCounter:
    _lock: threading.Lock
    _value: int

    def __init__(self, initial_value: int = 0):
        self._lock = threading.Lock()
        self._value = initial_value

    def inc_and_get(self) -> int:
        with self._lock:
            self._value += 1
            return self._value


def get_first_message_with_timeout(status_stream: SyncResponseIterator, timeout: int):
    waiter = future()

    def get_first_response(waiter):
        first_response = next(status_stream)
        waiter.set_result(first_response)

    thread = threading.Thread(
        target=get_first_response,
        args=(waiter,),
        name="first response attach stream thread",
        daemon=True,
    )
    thread.start()

    return waiter.result(timeout=timeout)



import asyncio
import concurrent.futures
import logging
import typing
from typing import List, Union, Optional

from ydb import issues
from ydb._grpc.grpcwrapper.common_utils import SupportedDriverType
from ydb._topic_common.common import (
    _get_shared_event_loop,
    CallFromSyncToAsync,
    TimeoutType,
)
from ydb._topic_reader import datatypes
from ydb._topic_reader.datatypes import PublicBatch
from ydb._topic_reader.topic_reader import (
    PublicReaderSettings,
    CommitResult,
)
from ydb._topic_reader.topic_reader_asyncio import (
    PublicAsyncIOReader,
    TopicReaderClosedError,
)

if typing.TYPE_CHECKING:
    from ..query.transaction import BaseQueryTxContext

logger = logging.getLogger(__name__)


class TopicReaderSync:
    _caller: CallFromSyncToAsync
    _async_reader: PublicAsyncIOReader
    _closed: bool
    _settings: PublicReaderSettings
    _parent: typing.Any  # need for prevent stop the client by GC

    def __init__(
        self,
        driver: SupportedDriverType,
        settings: PublicReaderSettings,
        *,
        eventloop: Optional[asyncio.AbstractEventLoop] = None,
        _parent=None,  # need for prevent stop the client by GC
    ):
        self._closed = False

        if eventloop:
            loop = eventloop
        else:
            loop = _get_shared_event_loop()

        self._caller = CallFromSyncToAsync(loop)

        async def create_reader():
            return PublicAsyncIOReader(driver, settings)

        self._async_reader = asyncio.run_coroutine_threadsafe(create_reader(), loop).result()

        self._settings = settings

        self._parent = _parent

    def __del__(self):
        if not self._closed:
            try:
                logger.debug("Topic reader was not closed properly. Consider using method close().")
                self.close(flush=False)
            except BaseException:
                logger.warning("Something went wrong during reader close in __del__")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def receive_message(self, *, timeout: TimeoutType = None) -> datatypes.PublicMessage:
        """
        Block until receive new message
        It has no async_ version for prevent lost messages, use async_wait_message as signal for new batches available.
        receive_message(timeout=0) may return None even right after async_wait_message() is ok - because lost of partition
        or connection to server lost

        if no new message in timeout seconds (default - infinite): raise TimeoutError()
        if timeout <= 0 - it will fast wait only one event loop cycle - without wait any i/o operations or pauses, get messages from internal buffer only.
        """
        self._check_closed()

        return self._caller.safe_call_with_result(self._async_reader.receive_message(), timeout)

    def async_wait_message(self) -> concurrent.futures.Future:
        """
        Returns a future, which will complete when the reader has at least one message in queue.
        If the reader already has a message - the future will complete immediately.

        A message may expire before it gets read so that the attempt to receive the message will fail
        despite the future has signaled about its availability.
        """
        self._check_closed()

        return self._caller.unsafe_call_with_future(self._async_reader.wait_message())

    def receive_batch(
        self,
        *,
        max_messages: typing.Union[int, None] = None,
        max_bytes: typing.Union[int, None] = None,
        timeout: Union[float, None] = None,
    ) -> Union[PublicBatch, None]:
        """
        Get one messages batch from reader
        It has no async_ version for prevent lost messages, use async_wait_message as signal for new batches available.

        if no new message in timeout seconds (default - infinite): raise TimeoutError()
        if timeout <= 0 - it will fast wait only one event loop cycle - without wait any i/o operations or pauses, get messages from internal buffer only.
        """
        self._check_closed()

        return self._caller.safe_call_with_result(
            self._async_reader.receive_batch(
                max_messages=max_messages,
            ),
            timeout,
        )

    def receive_batch_with_tx(
        self,
        tx: "BaseQueryTxContext",
        *,
        max_messages: typing.Union[int, None] = None,
        max_bytes: typing.Union[int, None] = None,
        timeout: Union[float, None] = None,
    ) -> Union[PublicBatch, None]:
        """
        Get one messages batch with tx from reader
        It has no async_ version for prevent lost messages, use async_wait_message as signal for new batches available.

        if no new message in timeout seconds (default - infinite): raise TimeoutError()
        if timeout <= 0 - it will fast wait only one event loop cycle - without wait any i/o operations or pauses, get messages from internal buffer only.
        """
        self._check_closed()

        return self._caller.safe_call_with_result(
            self._async_reader.receive_batch_with_tx(
                tx=tx,
                max_messages=max_messages,
            ),
            timeout,
        )

    def commit(self, mess: typing.Union[datatypes.PublicMessage, datatypes.PublicBatch]):
        """
        Put commit message to internal buffer.

        For the method no way check the commit result
        (for example if lost connection - commits will not re-send and committed messages will receive again)
        """
        self._check_closed()

        if self._settings.consumer is None:
            raise issues.Error("Commit operations are not supported for topic reader without consumer.")

        self._caller.call_sync(lambda: self._async_reader.commit(mess))

    def commit_with_ack(
        self,
        mess: typing.Union[datatypes.PublicMessage, datatypes.PublicBatch],
        timeout: TimeoutType = None,
    ) -> Union[CommitResult, List[CommitResult]]:
        """
        write commit message to a buffer and wait ack from the server.

        if receive in timeout seconds (default - infinite): raise TimeoutError()
        """
        self._check_closed()

        if self._settings.consumer is None:
            raise issues.Error("Commit operations are not supported for topic reader without consumer.")

        return self._caller.unsafe_call_with_result(self._async_reader.commit_with_ack(mess), timeout)

    def async_commit_with_ack(
        self, mess: typing.Union[datatypes.PublicMessage, datatypes.PublicBatch]
    ) -> concurrent.futures.Future:
        """
        write commit message to a buffer and return Future for wait result.
        """
        self._check_closed()

        if self._settings.consumer is None:
            raise issues.Error("Commit operations are not supported for topic reader without consumer.")

        return self._caller.unsafe_call_with_future(self._async_reader.commit_with_ack(mess))

    def close(self, *, flush: bool = True, timeout: TimeoutType = None):
        if self._closed:
            return

        self._closed = True

        self._caller.safe_call_with_result(self._async_reader.close(flush), timeout)

    def _check_closed(self):
        if self._closed:
            raise TopicReaderClosedError()

    @property
    def read_session_id(self) -> Optional[str]:
        return self._async_reader.read_session_id



from dataclasses import dataclass
import typing
from typing import Optional

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ..v4.protos import ydb_query_pb2
else:
    from ..common.protos import ydb_query_pb2

from . import ydb_query_public_types as public_types

from .common_utils import (
    IFromProto,
    IToProto,
    IFromPublic,
    ServerStatus,
)

from ... import convert


@dataclass
class CreateSessionResponse(IFromProto):
    status: ServerStatus
    session_id: str
    node_id: int

    @staticmethod
    def from_proto(msg: ydb_query_pb2.CreateSessionResponse) -> "CreateSessionResponse":
        return CreateSessionResponse(
            status=ServerStatus(msg.status, msg.issues),
            session_id=msg.session_id,
            node_id=msg.node_id,
        )


@dataclass
class DeleteSessionResponse(IFromProto):
    status: ServerStatus

    @staticmethod
    def from_proto(msg: ydb_query_pb2.DeleteSessionResponse) -> "DeleteSessionResponse":
        return DeleteSessionResponse(status=ServerStatus(msg.status, msg.issues))


@dataclass
class AttachSessionRequest(IToProto):
    session_id: str

    def to_proto(self) -> ydb_query_pb2.AttachSessionRequest:
        return ydb_query_pb2.AttachSessionRequest(session_id=self.session_id)


@dataclass
class TransactionMeta(IFromProto):
    tx_id: str

    @staticmethod
    def from_proto(msg: ydb_query_pb2.TransactionMeta) -> "TransactionMeta":
        return TransactionMeta(tx_id=msg.id)


@dataclass
class TransactionSettings(IFromPublic, IToProto):
    tx_mode: public_types.BaseQueryTxMode

    @staticmethod
    def from_public(tx_mode: public_types.BaseQueryTxMode) -> "TransactionSettings":
        return TransactionSettings(tx_mode=tx_mode)

    def to_proto(self) -> ydb_query_pb2.TransactionSettings:
        if self.tx_mode.name == "snapshot_read_write":
            return ydb_query_pb2.TransactionSettings(snapshot_read_write=self.tx_mode.to_proto())
        if self.tx_mode.name == "snapshot_read_only":
            return ydb_query_pb2.TransactionSettings(snapshot_read_only=self.tx_mode.to_proto())
        if self.tx_mode.name == "serializable_read_write":
            return ydb_query_pb2.TransactionSettings(serializable_read_write=self.tx_mode.to_proto())
        if self.tx_mode.name == "online_read_only":
            return ydb_query_pb2.TransactionSettings(online_read_only=self.tx_mode.to_proto())
        if self.tx_mode.name == "stale_read_only":
            return ydb_query_pb2.TransactionSettings(stale_read_only=self.tx_mode.to_proto())


@dataclass
class BeginTransactionRequest(IToProto):
    session_id: str
    tx_settings: TransactionSettings

    def to_proto(self) -> ydb_query_pb2.BeginTransactionRequest:
        return ydb_query_pb2.BeginTransactionRequest(
            session_id=self.session_id,
            tx_settings=self.tx_settings.to_proto(),
        )


@dataclass
class BeginTransactionResponse(IFromProto):
    status: Optional[ServerStatus]
    tx_meta: TransactionMeta

    @staticmethod
    def from_proto(msg: ydb_query_pb2.BeginTransactionResponse) -> "BeginTransactionResponse":
        return BeginTransactionResponse(
            status=ServerStatus(msg.status, msg.issues),
            tx_meta=TransactionMeta.from_proto(msg.tx_meta),
        )


@dataclass
class CommitTransactionResponse(IFromProto):
    status: Optional[ServerStatus]

    @staticmethod
    def from_proto(msg: ydb_query_pb2.CommitTransactionResponse) -> "CommitTransactionResponse":
        return CommitTransactionResponse(
            status=ServerStatus(msg.status, msg.issues),
        )


@dataclass
class RollbackTransactionResponse(IFromProto):
    status: Optional[ServerStatus]

    @staticmethod
    def from_proto(msg: ydb_query_pb2.RollbackTransactionResponse) -> "RollbackTransactionResponse":
        return RollbackTransactionResponse(
            status=ServerStatus(msg.status, msg.issues),
        )


@dataclass
class QueryContent(IFromPublic, IToProto):
    text: str
    syntax: int

    @staticmethod
    def from_public(query: str, syntax: int) -> "QueryContent":
        return QueryContent(text=query, syntax=syntax)

    def to_proto(self) -> ydb_query_pb2.QueryContent:
        return ydb_query_pb2.QueryContent(text=self.text, syntax=self.syntax)


@dataclass
class TransactionControl(IToProto):
    begin_tx: Optional[TransactionSettings]
    commit_tx: Optional[bool]
    tx_id: Optional[str]

    def to_proto(self) -> ydb_query_pb2.TransactionControl:
        if self.tx_id:
            return ydb_query_pb2.TransactionControl(
                tx_id=self.tx_id,
                commit_tx=self.commit_tx,
            )
        return ydb_query_pb2.TransactionControl(
            begin_tx=self.begin_tx.to_proto(),
            commit_tx=self.commit_tx,
        )


@dataclass
class ExecuteQueryRequest(IToProto):
    session_id: str
    query_content: QueryContent
    tx_control: TransactionControl
    concurrent_result_sets: bool
    exec_mode: int
    parameters: dict
    stats_mode: int
    schema_inclusion_mode: int
    result_set_format: int
    arrow_format_settings: Optional[public_types.ArrowFormatSettings]

    def to_proto(self) -> ydb_query_pb2.ExecuteQueryRequest:
        tx_control = self.tx_control.to_proto() if self.tx_control is not None else self.tx_control
        arrow_format_settings = (
            self.arrow_format_settings.to_proto() if self.arrow_format_settings is not None else None
        )
        return ydb_query_pb2.ExecuteQueryRequest(
            session_id=self.session_id,
            tx_control=tx_control,
            query_content=self.query_content.to_proto(),
            exec_mode=self.exec_mode,
            stats_mode=self.stats_mode,
            schema_inclusion_mode=self.schema_inclusion_mode,
            result_set_format=self.result_set_format,
            arrow_format_settings=arrow_format_settings,
            concurrent_result_sets=self.concurrent_result_sets,
            parameters=convert.query_parameters_to_pb(self.parameters),
        )



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_formats.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/ydb_formats.proto',
  package='Ydb.Formats',
  syntax='proto3',
  serialized_options=b'\n\026tech.ydb.proto.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18protos/ydb_formats.proto\x12\x0bYdb.Formats\"$\n\x12\x41rrowBatchSettings\x12\x0e\n\x06schema\x18\x01 \x01(\x0c\"\xda\x01\n\x0b\x43svSettings\x12\x11\n\tskip_rows\x18\x01 \x01(\r\x12\x11\n\tdelimiter\x18\x02 \x01(\x0c\x12\x12\n\nnull_value\x18\x03 \x01(\x0c\x12\x0e\n\x06header\x18\x04 \x01(\x08\x12\x31\n\x07quoting\x18\x05 \x01(\x0b\x32 .Ydb.Formats.CsvSettings.Quoting\x1aN\n\x07Quoting\x12\x10\n\x08\x64isabled\x18\x01 \x01(\x08\x12\x12\n\nquote_char\x18\x02 \x01(\x0c\x12\x1d\n\x15\x64ouble_quote_disabled\x18\x03 \x01(\x08\"\xac\x02\n\x13\x41rrowFormatSettings\x12L\n\x11\x63ompression_codec\x18\x01 \x01(\x0b\x32\x31.Ydb.Formats.ArrowFormatSettings.CompressionCodec\x1a\xc6\x01\n\x10\x43ompressionCodec\x12\x44\n\x04type\x18\x01 \x01(\x0e\x32\x36.Ydb.Formats.ArrowFormatSettings.CompressionCodec.Type\x12\x12\n\x05level\x18\x02 \x01(\x05H\x00\x88\x01\x01\"N\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\r\n\tTYPE_NONE\x10\x01\x12\r\n\tTYPE_ZSTD\x10\x02\x12\x12\n\x0eTYPE_LZ4_FRAME\x10\x03\x42\x08\n\x06_level\"!\n\x0f\x41rrowFormatMeta\x12\x0e\n\x06schema\x18\x01 \x01(\x0c\x42W\n\x16tech.ydb.proto.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\xf8\x01\x01\x62\x06proto3'
)



_ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='Ydb.Formats.ArrowFormatSettings.CompressionCodec.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_NONE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_ZSTD', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TYPE_LZ4_FRAME', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=513,
  serialized_end=591,
)
_sym_db.RegisterEnumDescriptor(_ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE)


_ARROWBATCHSETTINGS = _descriptor.Descriptor(
  name='ArrowBatchSettings',
  full_name='Ydb.Formats.ArrowBatchSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='Ydb.Formats.ArrowBatchSettings.schema', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=77,
)


_CSVSETTINGS_QUOTING = _descriptor.Descriptor(
  name='Quoting',
  full_name='Ydb.Formats.CsvSettings.Quoting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='disabled', full_name='Ydb.Formats.CsvSettings.Quoting.disabled', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='quote_char', full_name='Ydb.Formats.CsvSettings.Quoting.quote_char', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double_quote_disabled', full_name='Ydb.Formats.CsvSettings.Quoting.double_quote_disabled', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=220,
  serialized_end=298,
)

_CSVSETTINGS = _descriptor.Descriptor(
  name='CsvSettings',
  full_name='Ydb.Formats.CsvSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='skip_rows', full_name='Ydb.Formats.CsvSettings.skip_rows', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delimiter', full_name='Ydb.Formats.CsvSettings.delimiter', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_value', full_name='Ydb.Formats.CsvSettings.null_value', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='header', full_name='Ydb.Formats.CsvSettings.header', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='quoting', full_name='Ydb.Formats.CsvSettings.quoting', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CSVSETTINGS_QUOTING, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=298,
)


_ARROWFORMATSETTINGS_COMPRESSIONCODEC = _descriptor.Descriptor(
  name='CompressionCodec',
  full_name='Ydb.Formats.ArrowFormatSettings.CompressionCodec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='Ydb.Formats.ArrowFormatSettings.CompressionCodec.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='Ydb.Formats.ArrowFormatSettings.CompressionCodec.level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_level', full_name='Ydb.Formats.ArrowFormatSettings.CompressionCodec._level',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=403,
  serialized_end=601,
)

_ARROWFORMATSETTINGS = _descriptor.Descriptor(
  name='ArrowFormatSettings',
  full_name='Ydb.Formats.ArrowFormatSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='compression_codec', full_name='Ydb.Formats.ArrowFormatSettings.compression_codec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ARROWFORMATSETTINGS_COMPRESSIONCODEC, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=301,
  serialized_end=601,
)


_ARROWFORMATMETA = _descriptor.Descriptor(
  name='ArrowFormatMeta',
  full_name='Ydb.Formats.ArrowFormatMeta',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='Ydb.Formats.ArrowFormatMeta.schema', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=603,
  serialized_end=636,
)

_CSVSETTINGS_QUOTING.containing_type = _CSVSETTINGS
_CSVSETTINGS.fields_by_name['quoting'].message_type = _CSVSETTINGS_QUOTING
_ARROWFORMATSETTINGS_COMPRESSIONCODEC.fields_by_name['type'].enum_type = _ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE
_ARROWFORMATSETTINGS_COMPRESSIONCODEC.containing_type = _ARROWFORMATSETTINGS
_ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE.containing_type = _ARROWFORMATSETTINGS_COMPRESSIONCODEC
_ARROWFORMATSETTINGS_COMPRESSIONCODEC.oneofs_by_name['_level'].fields.append(
  _ARROWFORMATSETTINGS_COMPRESSIONCODEC.fields_by_name['level'])
_ARROWFORMATSETTINGS_COMPRESSIONCODEC.fields_by_name['level'].containing_oneof = _ARROWFORMATSETTINGS_COMPRESSIONCODEC.oneofs_by_name['_level']
_ARROWFORMATSETTINGS.fields_by_name['compression_codec'].message_type = _ARROWFORMATSETTINGS_COMPRESSIONCODEC
DESCRIPTOR.message_types_by_name['ArrowBatchSettings'] = _ARROWBATCHSETTINGS
DESCRIPTOR.message_types_by_name['CsvSettings'] = _CSVSETTINGS
DESCRIPTOR.message_types_by_name['ArrowFormatSettings'] = _ARROWFORMATSETTINGS
DESCRIPTOR.message_types_by_name['ArrowFormatMeta'] = _ARROWFORMATMETA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ArrowBatchSettings = _reflection.GeneratedProtocolMessageType('ArrowBatchSettings', (_message.Message,), {
  'DESCRIPTOR' : _ARROWBATCHSETTINGS,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.ArrowBatchSettings)
  })
_sym_db.RegisterMessage(ArrowBatchSettings)

CsvSettings = _reflection.GeneratedProtocolMessageType('CsvSettings', (_message.Message,), {

  'Quoting' : _reflection.GeneratedProtocolMessageType('Quoting', (_message.Message,), {
    'DESCRIPTOR' : _CSVSETTINGS_QUOTING,
    '__module__' : 'protos.ydb_formats_pb2'
    # @@protoc_insertion_point(class_scope:Ydb.Formats.CsvSettings.Quoting)
    })
  ,
  'DESCRIPTOR' : _CSVSETTINGS,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.CsvSettings)
  })
_sym_db.RegisterMessage(CsvSettings)
_sym_db.RegisterMessage(CsvSettings.Quoting)

ArrowFormatSettings = _reflection.GeneratedProtocolMessageType('ArrowFormatSettings', (_message.Message,), {

  'CompressionCodec' : _reflection.GeneratedProtocolMessageType('CompressionCodec', (_message.Message,), {
    'DESCRIPTOR' : _ARROWFORMATSETTINGS_COMPRESSIONCODEC,
    '__module__' : 'protos.ydb_formats_pb2'
    # @@protoc_insertion_point(class_scope:Ydb.Formats.ArrowFormatSettings.CompressionCodec)
    })
  ,
  'DESCRIPTOR' : _ARROWFORMATSETTINGS,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.ArrowFormatSettings)
  })
_sym_db.RegisterMessage(ArrowFormatSettings)
_sym_db.RegisterMessage(ArrowFormatSettings.CompressionCodec)

ArrowFormatMeta = _reflection.GeneratedProtocolMessageType('ArrowFormatMeta', (_message.Message,), {
  'DESCRIPTOR' : _ARROWFORMATMETA,
  '__module__' : 'protos.ydb_formats_pb2'
  # @@protoc_insertion_point(class_scope:Ydb.Formats.ArrowFormatMeta)
  })
_sym_db.RegisterMessage(ArrowFormatMeta)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
import time
import http.server
import urllib
import json
import threading
import tempfile
import os
import jwt
import base64
from ydb.oauth2_token_exchange import Oauth2TokenExchangeCredentials, FixedTokenSource
from ydb.driver import credentials_from_env_variables


TEST_RSA_PRIVATE_KEY_CONTENT = "-----BEGIN PRIVATE KEY-----\nMIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQC75/JS3rMcLJxv\nFgpOzF5+2gH+Yig3RE2MTl9uwC0BZKAv6foYr7xywQyWIK+W1cBhz8R4LfFmZo2j\nM0aCvdRmNBdW0EDSTnHLxCsFhoQWLVq+bI5f5jzkcoiioUtaEpADPqwgVULVtN/n\nnPJiZ6/dU30C3jmR6+LUgEntUtWt3eq3xQIn5lG3zC1klBY/HxtfH5Hu8xBvwRQT\nJnh3UpPLj8XwSmriDgdrhR7o6umWyVuGrMKlLHmeivlfzjYtfzO1MOIMG8t2/zxG\nR+xb4Vwks73sH1KruH/0/JMXU97npwpe+Um+uXhpldPygGErEia7abyZB2gMpXqr\nWYKMo02NAgMBAAECggEAO0BpC5OYw/4XN/optu4/r91bupTGHKNHlsIR2rDzoBhU\nYLd1evpTQJY6O07EP5pYZx9mUwUdtU4KRJeDGO/1/WJYp7HUdtxwirHpZP0lQn77\nuccuX/QQaHLrPekBgz4ONk+5ZBqukAfQgM7fKYOLk41jgpeDbM2Ggb6QUSsJISEp\nzrwpI/nNT/wn+Hvx4DxrzWU6wF+P8kl77UwPYlTA7GsT+T7eKGVH8xsxmK8pt6lg\nsvlBA5XosWBWUCGLgcBkAY5e4ZWbkdd183o+oMo78id6C+PQPE66PLDtHWfpRRmN\nm6XC03x6NVhnfvfozoWnmS4+e4qj4F/emCHvn0GMywKBgQDLXlj7YPFVXxZpUvg/\nrheVcCTGbNmQJ+4cZXx87huqwqKgkmtOyeWsRc7zYInYgraDrtCuDBCfP//ZzOh0\nLxepYLTPk5eNn/GT+VVrqsy35Ccr60g7Lp/bzb1WxyhcLbo0KX7/6jl0lP+VKtdv\nmto+4mbSBXSM1Y5BVVoVgJ3T/wKBgQDsiSvPRzVi5TTj13x67PFymTMx3HCe2WzH\nJUyepCmVhTm482zW95pv6raDr5CTO6OYpHtc5sTTRhVYEZoEYFTM9Vw8faBtluWG\nBjkRh4cIpoIARMn74YZKj0C/0vdX7SHdyBOU3bgRPHg08Hwu3xReqT1kEPSI/B2V\n4pe5fVrucwKBgQCNFgUxUA3dJjyMES18MDDYUZaRug4tfiYouRdmLGIxUxozv6CG\nZnbZzwxFt+GpvPUV4f+P33rgoCvFU+yoPctyjE6j+0aW0DFucPmb2kBwCu5J/856\nkFwCx3blbwFHAco+SdN7g2kcwgmV2MTg/lMOcU7XwUUcN0Obe7UlWbckzQKBgQDQ\nnXaXHL24GGFaZe4y2JFmujmNy1dEsoye44W9ERpf9h1fwsoGmmCKPp90az5+rIXw\nFXl8CUgk8lXW08db/r4r+ma8Lyx0GzcZyplAnaB5/6j+pazjSxfO4KOBy4Y89Tb+\nTP0AOcCi6ws13bgY+sUTa/5qKA4UVw+c5zlb7nRpgwKBgGXAXhenFw1666482iiN\ncHSgwc4ZHa1oL6aNJR1XWH+aboBSwR+feKHUPeT4jHgzRGo/aCNHD2FE5I8eBv33\nof1kWYjAO0YdzeKrW0rTwfvt9gGg+CS397aWu4cy+mTI+MNfBgeDAIVBeJOJXLlX\nhL8bFAuNNVrCOp79TNnNIsh7\n-----END PRIVATE KEY-----\n"  # noqa: E501
TEST_RSA_PUBLIC_KEY_CONTENT = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAu+fyUt6zHCycbxYKTsxe\nftoB/mIoN0RNjE5fbsAtAWSgL+n6GK+8csEMliCvltXAYc/EeC3xZmaNozNGgr3U\nZjQXVtBA0k5xy8QrBYaEFi1avmyOX+Y85HKIoqFLWhKQAz6sIFVC1bTf55zyYmev\n3VN9At45kevi1IBJ7VLVrd3qt8UCJ+ZRt8wtZJQWPx8bXx+R7vMQb8EUEyZ4d1KT\ny4/F8Epq4g4Ha4Ue6OrplslbhqzCpSx5nor5X842LX8ztTDiDBvLdv88RkfsW+Fc\nJLO97B9Sq7h/9PyTF1Pe56cKXvlJvrl4aZXT8oBhKxImu2m8mQdoDKV6q1mCjKNN\njQIDAQAB\n-----END PUBLIC KEY-----\n"  # noqa: E501
TEST_EC_PRIVATE_KEY_CONTENT = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIB6fv25gf7P/7fkjW/2kcKICUhHeOygkFeUJ/ylyU3hloAoGCCqGSM49\nAwEHoUQDQgAEvkKy92hpLiT0GEpzFkYBEWWnkAGTTA6141H0oInA9X30eS0RObAa\nmVY8yD39NI7Nj03hBxEa4Z0tOhrq9cW8eg==\n-----END EC PRIVATE KEY-----\n"  # noqa: E501
TEST_EC_PUBLIC_KEY_CONTENT = "-----BEGIN PUBLIC KEY-----\nMFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEvkKy92hpLiT0GEpzFkYBEWWnkAGT\nTA6141H0oInA9X30eS0RObAamVY8yD39NI7Nj03hBxEa4Z0tOhrq9cW8eg==\n-----END PUBLIC KEY-----\n"  # noqa: E501
TEST_HMAC_SECRET_KEY_BASE64_CONTENT = "VGhlIHdvcmxkIGhhcyBjaGFuZ2VkLgpJIHNlZSBpdCBpbiB0aGUgd2F0ZXIuCkkgZmVlbCBpdCBpbiB0aGUgRWFydGguCkkgc21lbGwgaXQgaW4gdGhlIGFpci4KTXVjaCB0aGF0IG9uY2Ugd2FzIGlzIGxvc3QsCkZvciBub25lIG5vdyBsaXZlIHdobyByZW1lbWJlciBpdC4K"  # noqa: E501


def get_expire_time(creds):
    return creds._expires_in - time.time()


TOKEN_EXCHANGE_RESPONSES = {
    "test_src_token": [
        200,
        {
            "access_token": "test_dst_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "BeareR",
            "expires_in": 42,
        },
    ],
    "test_http_code": [
        403,
        {
            "err": "err_42",
            "access_token": "test_dst_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "BeareR",
            "expires_in": 42,
        },
    ],
    "test_wrong_token_type": [
        200,
        {
            "access_token": "test_dst_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "basic",
            "expires_in": 42,
        },
    ],
    "test_wrong_expiration_time": [
        200,
        {
            "access_token": "test_dst_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "BeareR",
            "expires_in": -42,
        },
    ],
    "test_different_scope": [
        200,
        {
            "access_token": "test_dst_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "BeareR",
            "expires_in": 42,
            "scope": "s1",
        },
    ],
}


class FixedTokenSourceChecker:
    def __init__(
        self,
        token,
        token_type,
    ):
        self.token = token
        self.token_type = token_type

    def check(self, token, token_type):
        assert token == self.token
        assert token_type == self.token_type


class AnyTokenSourceChecker:
    def check(self, token, token_type):
        assert token != ""
        assert token_type != ""


class JwtTokenSourceChecker:
    def __init__(
        self,
        alg,
        public_key,
        key_id=None,
        issuer=None,
        subject=None,
        aud=None,
        id=None,
        ttl_seconds=3600,
    ):
        self.alg = alg
        self.public_key = public_key
        self.key_id = key_id
        self.issuer = issuer
        self.subject = subject
        self.aud = aud
        self.id = id
        self.ttl_seconds = ttl_seconds

    def check(self, token, token_type):
        assert token_type == "urn:ietf:params:oauth:token-type:jwt"
        decoded = jwt.decode(
            token,
            key=self.public_key,
            algorithms=[self.alg],
            options={
                "require": ["iat", "exp"],
                "verify_signature": True,
                "verify_aud": False,
                "verify_iss": False,
            },
        )
        header = jwt.get_unverified_header(token)
        assert header.get("kid") == self.key_id
        assert header.get("alg") == self.alg
        assert decoded.get("iss") == self.issuer
        assert decoded.get("sub") == self.subject
        assert decoded.get("aud") == self.aud
        assert decoded.get("jti") == self.id
        assert abs(decoded["iat"] - time.time()) <= 60
        assert abs(decoded["exp"] - decoded["iat"]) == self.ttl_seconds


class Oauth2ExchangeServiceChecker:
    def __init__(
        self,
        subject_token_source=None,
        actor_token_source=None,
        audience=None,
        scope=None,
        resource=None,
        requested_token_type="urn:ietf:params:oauth:token-type:access_token",
        grant_type="urn:ietf:params:oauth:grant-type:token-exchange",
    ):
        self.subject_token_source = subject_token_source
        self.actor_token_source = actor_token_source
        self.audience = audience
        self.scope = scope
        self.resource = resource
        self.requested_token_type = requested_token_type
        self.grant_type = grant_type

    def check(self, handler, parsed_request) -> None:
        assert handler.headers["Content-Type"] == "application/x-www-form-urlencoded"
        assert handler.path == "/token/exchange"

        assert len(parsed_request["grant_type"]) == 1
        assert parsed_request["grant_type"][0] == self.grant_type

        assert len(parsed_request["requested_token_type"]) == 1
        assert parsed_request["requested_token_type"][0] == self.requested_token_type

        if self.audience is None or len(self.audience) == 0:
            assert len(parsed_request.get("audience", [])) == 0
        else:
            assert len(parsed_request["audience"]) == len(self.audience)
            for i in range(len(self.audience)):
                assert parsed_request["audience"][i] == self.audience[i]

        if self.scope is None or len(self.scope) == 0:
            assert len(parsed_request.get("scope", [])) == 0
        else:
            assert len(parsed_request.get("scope", [])) == 1
            assert parsed_request["scope"][0] == " ".join(self.scope)

        if self.resource is None or len(self.resource) == 0:
            assert len(parsed_request.get("resource", [])) == 0
        else:
            assert len(parsed_request["resource"]) == len(self.resource)
            for i in range(len(self.resource)):
                assert parsed_request["resource"][i] == self.resource[i]

        if self.subject_token_source is None:
            assert len(parsed_request.get("subject_token", [])) == 0
            assert len(parsed_request.get("subject_token_type", [])) == 0
        else:
            assert len(parsed_request.get("subject_token", [])) == 1
            assert len(parsed_request.get("subject_token_type", [])) == 1
            self.subject_token_source.check(parsed_request["subject_token"][0], parsed_request["subject_token_type"][0])

        if self.actor_token_source is None:
            assert len(parsed_request.get("actor_token", [])) == 0
            assert len(parsed_request.get("actor_token_type", [])) == 0
        else:
            assert len(parsed_request.get("actor_token", [])) == 1
            assert len(parsed_request.get("actor_token_type", [])) == 1
            self.actor_token_source.check(parsed_request["actor_token"][0], parsed_request["actor_token_type"][0])


class TokenExchangeResponseBySubjectToken:
    def __init__(self, responses=TOKEN_EXCHANGE_RESPONSES):
        self.responses = responses

    def get_response(self, parsed_request):
        assert len(parsed_request["subject_token"]) == 1
        responses = self.responses.get(parsed_request["subject_token"][0])
        assert responses is not None
        response_code = responses[0]
        response = responses[1]
        return response_code, response


class Oauth2TokenExchangeResponse:
    def __init__(
        self,
        response_code,
        response,
    ):
        self.response_code = response_code
        self.response = response

    def get_response(self, parsed_request):
        return self.response_code, self.response


class OauthExchangeServiceHandler(http.server.BaseHTTPRequestHandler):
    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length).decode("utf8")
        print("OauthExchangeServiceHandler.POST data: {}".format(post_data))
        parsed_request = urllib.parse.parse_qs(str(post_data))
        try:
            self.server.checker.check(self, parsed_request)
            self.server.checker.check_successful = True
            response_code, response = self.server.response.get_response(parsed_request)

            self.send_response(response_code)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write(json.dumps(response).encode("utf8"))
        except Exception as ex:
            self.send_response(500)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write("Exception during text check: {}".format(ex).encode("utf8"))


class Oauth2TokenExchangeServiceForTest(http.server.HTTPServer):
    def __init__(self, port, checker=None, response=None):
        http.server.HTTPServer.__init__(self, ("localhost", port), OauthExchangeServiceHandler)
        self.checker = checker
        if self.checker is None:
            self.checker = Oauth2ExchangeServiceChecker(
                subject_token_source=AnyTokenSourceChecker(),
                audience=["a1", "a2"],
                scope=["s1", "s2"],
            )
        self.response = response
        if self.response is None:
            self.response = TokenExchangeResponseBySubjectToken()
        self.port = port

    def endpoint(self):
        return "http://localhost:{}/token/exchange".format(self.port)


class DataForTest:
    def __init__(self, src_token, dst_token, error_text):
        self.src_token = src_token
        self.dst_token = dst_token
        self.error_text = error_text

    def run_check(self, server):
        try:
            credentials = Oauth2TokenExchangeCredentials(
                server.endpoint(),
                subject_token_source=FixedTokenSource(self.src_token, "test_token_type"),
                audience=["a1", "a2"],
                scope=["s1", "s2"],
            )
            t = credentials.get_auth_token()
            assert not self.error_text, "Exception is expected. Test: {}".format(self.src_token)
            assert t == self.dst_token
            assert get_expire_time(credentials) <= 42
        except AssertionError:
            raise
        except Exception as ex:
            if self.error_text:
                assert self.error_text in str(ex)
            else:
                assert False, "Exception is not expected. Test: {}. Exception text: {}".format(self.src_token, ex)


def test_oauth2_token_exchange_credentials():
    server = Oauth2TokenExchangeServiceForTest(40123)

    tests = [
        DataForTest(
            "test_src_token",
            "Bearer test_dst_token",
            "",
        ),
        DataForTest(
            "test_http_code",
            "",
            "err_42",
        ),
        DataForTest(
            "test_wrong_token_type",
            "",
            "unsupported token type: basic",
        ),
        DataForTest(
            "test_wrong_expiration_time",
            "",
            "incorrect expiration time: -42",
        ),
        DataForTest(
            "test_different_scope",
            "",
            'different scope. Expected: "s1 s2", but got: "s1"',
        ),
    ]

    def serve(s):
        for _ in tests:
            # one request per test
            s.handle_request()

    serve_thread = threading.Thread(target=serve, args=(server,))
    serve_thread.start()

    try:
        for t in tests:
            t.run_check(server)
    except Exception:
        raise

    serve_thread.join()


class DataForConfigTest:
    def __init__(
        self,
        cfg=None,  # cfg or cfg text
        cfg_file=None,
        checker=None,
        response=None,
        init_error_text_part=None,
        get_token_error_text_part=None,
        dst_token=None,
        dst_expire_time=42,
        http_request_is_expected=None,
        init_from_env=False,
    ):
        self.cfg = cfg
        self.cfg_file = cfg_file
        self.checker = checker
        self.response = response
        self.init_error_text_part = init_error_text_part
        self.get_token_error_text_part = get_token_error_text_part
        self.dst_token = dst_token
        self.dst_expire_time = dst_expire_time
        self.http_request_is_expected = http_request_is_expected
        self.init_from_env = init_from_env

    def get_cfg(self):
        if isinstance(self.cfg, str):
            return self.cfg
        else:
            return json.dumps(self.cfg, indent=4)

    def expect_http_request(self):
        if self.http_request_is_expected is not None:
            return self.http_request_is_expected

        if self.init_error_text_part is not None:
            return False
        return True

    def run_check(self, server):
        server.checker = self.checker
        if server.checker is not None:
            server.checker.check_successful = False
        server.response = self.response

        if self.cfg_file:
            cfg_file = self.cfg_file

            def rm_file():
                pass

        else:
            temp_cfg_file = tempfile.NamedTemporaryFile(delete=False)
            cfg_file = temp_cfg_file.name
            temp_cfg_file.write(self.get_cfg().encode("utf-8"))
            temp_cfg_file.close()

            def rm_file():
                os.remove(cfg_file)
                if self.init_from_env:
                    del os.environ["YDB_OAUTH2_KEY_FILE"]

        try:
            if self.init_from_env:
                os.environ["YDB_OAUTH2_KEY_FILE"] = cfg_file
                credentials = credentials_from_env_variables()
            else:
                credentials = Oauth2TokenExchangeCredentials.from_file(
                    cfg_file,
                    iam_endpoint=server.endpoint(),
                )
            assert self.init_error_text_part is None, "Init exception is expected. Test:\n{}".format(self.get_cfg())

            t = credentials.get_auth_token()
            assert not self.get_token_error_text_part, "Exception is expected. Test:\n{}".format(self.get_cfg())
            if self.expect_http_request() and server.checker is not None:
                assert server.checker.check_successful

            assert t == self.dst_token
            assert get_expire_time(credentials) <= self.dst_expire_time
            rm_file()
        except AssertionError:
            rm_file()
            raise
        except Exception as ex:
            rm_file()
            err_text = self.init_error_text_part if self.init_error_text_part else self.get_token_error_text_part
            if err_text:
                assert err_text in str(ex)
            else:
                assert False, "Exception is not expected. Test:\n{}. Exception text: {}".format(self.get_cfg(), ex)


def test_oauth2_token_exchange_credentials_file():
    server = Oauth2TokenExchangeServiceForTest(40124)

    tests = [
        DataForConfigTest(cfg="not json config", init_error_text_part="Failed to parse json config"),
        DataForConfigTest(
            init_from_env=True,
            cfg={
                "res": "tEst",
                "grant-type": "grant",
                "requested-token-type": "access_token",
                "subject-credentials": {
                    "type": "fixed",
                    "token": "test-token",
                    "token-type": "test-token-type",
                },
            },
            init_error_text_part="no token endpoint specified",
        ),
        DataForConfigTest(
            init_from_env=True,
            cfg={
                "token-endpoint": server.endpoint(),
                "res": "tEst",
                "grant-type": "grant",
                "requested-token-type": "access_token",
                "subject-credentials": {
                    "type": "fixed",
                    "token": "test-token",
                    "token-type": "test-token-type",
                },
            },
            checker=Oauth2ExchangeServiceChecker(
                subject_token_source=FixedTokenSourceChecker(
                    token="test-token",
                    token_type="test-token-type",
                ),
                grant_type="grant",
                requested_token_type="access_token",
                resource=["tEst"],
            ),
            response=Oauth2TokenExchangeResponse(
                200,
                {
                    "access_token": "test_dst_token",
                    "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "token_type": "Bearer",
                    "expires_in": 42,
                },
            ),
            dst_token="Bearer test_dst_token",
        ),
        DataForConfigTest(
            cfg={
                "aud": "test-aud",
                "scope": [
                    "s1",
                    "s2",
                ],
                "res": ["r1", "r2"],
                "unknown-field": [123],
                "actor-credentials": {
                    "type": "fixed",
                    "token": "test-token",
                    "token-type": "test-token-type",
                },
            },
            checker=Oauth2ExchangeServiceChecker(
                actor_token_source=FixedTokenSourceChecker(
                    token="test-token",
                    token_type="test-token-type",
                ),
                audience=["test-aud"],
                scope=["s1", "s2"],
                resource=["r1", "r2"],
            ),
            response=Oauth2TokenExchangeResponse(
                200,
                {
                    "access_token": "test_dst_token",
                    "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "token_type": "Bearer",
                    "expires_in": 42,
                },
            ),
            dst_token="Bearer test_dst_token",
        ),
        DataForConfigTest(
            cfg={
                "requested-token-type": "access_token",
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "ps256",
                    "private-key": TEST_RSA_PRIVATE_KEY_CONTENT,
                    "aud": ["a1", "a2"],
                    "jti": "123",
                    "sub": "test_subject",
                    "iss": "test_issuer",
                    "kid": "test_key_id",
                    "ttl": "24h",
                    "unknown_field": "hello!",
                },
            },
            checker=Oauth2ExchangeServiceChecker(
                subject_token_source=JwtTokenSourceChecker(
                    alg="PS256",
                    public_key=TEST_RSA_PUBLIC_KEY_CONTENT,
                    aud=["a1", "a2"],
                    id="123",
                    subject="test_subject",
                    issuer="test_issuer",
                    key_id="test_key_id",
                    ttl_seconds=24 * 3600,
                ),
                requested_token_type="access_token",
            ),
            response=Oauth2TokenExchangeResponse(
                200,
                {
                    "access_token": "test_dst_token",
                    "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "token_type": "bearer",
                    "expires_in": 42,
                },
            ),
            dst_token="Bearer test_dst_token",
        ),
        DataForConfigTest(
            cfg={
                "actor-credentials": {
                    "type": "JWT",
                    "alg": "es256",
                    "private-key": TEST_EC_PRIVATE_KEY_CONTENT,
                    "ttl": "3m",
                    "unknown_field": "hello!",
                },
            },
            checker=Oauth2ExchangeServiceChecker(
                actor_token_source=JwtTokenSourceChecker(
                    alg="ES256",
                    public_key=TEST_EC_PUBLIC_KEY_CONTENT,
                    ttl_seconds=180,
                ),
            ),
            response=Oauth2TokenExchangeResponse(
                200,
                {
                    "access_token": "test_dst_token",
                    "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "token_type": "bearer",
                    "expires_in": 42,
                },
            ),
            dst_token="Bearer test_dst_token",
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "hs512",
                    "private-key": TEST_HMAC_SECRET_KEY_BASE64_CONTENT,
                },
            },
            checker=Oauth2ExchangeServiceChecker(
                subject_token_source=JwtTokenSourceChecker(
                    alg="HS512",
                    public_key=base64.b64decode(TEST_HMAC_SECRET_KEY_BASE64_CONTENT),
                ),
            ),
            response=Oauth2TokenExchangeResponse(
                200,
                {
                    "access_token": "test_dst_token",
                    "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                    "token_type": "bearer",
                    "expires_in": 42,
                },
            ),
            dst_token="Bearer test_dst_token",
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "rs512",
                    "private-key": TEST_HMAC_SECRET_KEY_BASE64_CONTENT,
                },
            },
            http_request_is_expected=False,
            get_token_error_text_part="Could not deserialize key data.",
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "es512",
                    "private-key": TEST_HMAC_SECRET_KEY_BASE64_CONTENT,
                },
            },
            http_request_is_expected=False,
            get_token_error_text_part="Could not deserialize key data.",
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "es512",
                    "private-key": TEST_RSA_PRIVATE_KEY_CONTENT,
                },
            },
            http_request_is_expected=False,
            get_token_error_text_part="sign() missing 1 required positional argument",
        ),
        DataForConfigTest(
            cfg_file="~/unknown-file.cfg",
            init_error_text_part="No such file or directory",
        ),
        DataForConfigTest(
            cfg={
                "actor-credentials": "",
            },
            init_error_text_part='Key "actor-credentials" is expected to be a json map',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "RS256",
                    "private-key": "123",
                    "ttl": 123,
                },
            },
            init_error_text_part='Key "ttl" is expected to be a string',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "RS256",
                    "private-key": "123",
                    "ttl": "-3h",
                },
            },
            init_error_text_part="-3: negative duration is not allowed",
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "private-key": "123",
                },
            },
            init_error_text_part='Key "alg" is expected to be a nonempty string',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "HS384",
                },
            },
            init_error_text_part='Key "private-key" is expected to be a nonempty string',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "unknown",
                    "private-key": "123",
                },
            },
            http_request_is_expected=False,
            get_token_error_text_part="Algorithm not supported.",
        ),
        DataForConfigTest(
            cfg={
                "aud": {
                    "value": "wrong_format of aud: not string and not list",
                },
                "subject-credentials": {
                    "type": "FIXED",
                    "token": "test-token",
                    "token-type": "test-token-type",
                },
            },
            init_error_text_part='Key "aud" is expected to be a single string or list of nonempty strings',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "JWT",
                    "alg": "RS256",
                    "private-key": "123",
                    "aud": {
                        "value": "wrong_format of aud: not string and not list",
                    },
                },
            },
            init_error_text_part='Key "aud" is expected to be a single string or list of nonempty strings',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "unknown",
                },
            },
            init_error_text_part='"subject-credentials": unknown token source type: "unknown"',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "token": "test",
                },
            },
            init_error_text_part='Key "type" is expected to be a nonempty string',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "FIXED",
                    "token": "test",
                },
            },
            init_error_text_part='Key "token-type" is expected to be a nonempty string',
        ),
        DataForConfigTest(
            cfg={
                "subject-credentials": {
                    "type": "FIXED",
                    "token-type": "test",
                },
            },
            init_error_text_part='Key "token" is expected to be a nonempty string',
        ),
    ]

    def serve(s):
        for t in tests:
            # one request per test
            if t.expect_http_request():
                s.handle_request()

    serve_thread = threading.Thread(target=serve, args=(server,))
    serve_thread.start()

    for t in tests:
        t.run_check(server)

    serve_thread.join()



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ydb_monitoring_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'ydb_monitoring_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_monitoring_pb2 as protos_dot_ydb__monitoring__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17ydb_monitoring_v1.proto\x12\x11Ydb.Monitoring.V1\x1a\x1bprotos/ydb_monitoring.proto2\xb7\x01\n\x11MonitoringService\x12P\n\tSelfCheck\x12 .Ydb.Monitoring.SelfCheckRequest\x1a!.Ydb.Monitoring.SelfCheckResponse\x12P\n\tNodeCheck\x12 .Ydb.Monitoring.NodeCheckRequest\x1a!.Ydb.Monitoring.NodeCheckResponseBY\n\x1ctech.ydb.proto.monitoring.v1Z9github.com/ydb-platform/ydb-go-genproto/Ydb_Monitoring_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_monitoring_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034tech.ydb.proto.monitoring.v1Z9github.com/ydb-platform/ydb-go-genproto/Ydb_Monitoring_V1'
  _globals['_MONITORINGSERVICE']._serialized_start=76
  _globals['_MONITORINGSERVICE']._serialized_end=259
# @@protoc_insertion_point(module_scope)



from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_export_pb2 as _ydb_export_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ImportDataRequest(_message.Message):
    __slots__ = ["data", "operation_params", "path", "ydb_dump"]
    DATA_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    YDB_DUMP_FIELD_NUMBER: _ClassVar[int]
    data: bytes
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    ydb_dump: YdbDumpFormat
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., data: _Optional[bytes] = ..., ydb_dump: _Optional[_Union[YdbDumpFormat, _Mapping]] = ...) -> None: ...

class ImportDataResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ImportDataResult(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ImportFromFsMetadata(_message.Message):
    __slots__ = ["items_progress", "progress", "settings"]
    ITEMS_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    items_progress: _containers.RepeatedCompositeFieldContainer[ImportItemProgress]
    progress: ImportProgress.Progress
    settings: ImportFromFsSettings
    def __init__(self, settings: _Optional[_Union[ImportFromFsSettings, _Mapping]] = ..., progress: _Optional[_Union[ImportProgress.Progress, str]] = ..., items_progress: _Optional[_Iterable[_Union[ImportItemProgress, _Mapping]]] = ...) -> None: ...

class ImportFromFsRequest(_message.Message):
    __slots__ = ["operation_params", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    settings: ImportFromFsSettings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ImportFromFsSettings, _Mapping]] = ...) -> None: ...

class ImportFromFsResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ImportFromFsResult(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ImportFromFsSettings(_message.Message):
    __slots__ = ["base_path", "description", "destination_path", "encryption_settings", "exclude_regexps", "items", "no_acl", "number_of_retries", "skip_checksum_validation"]
    class Item(_message.Message):
        __slots__ = ["destination_path", "source_path"]
        DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
        destination_path: str
        source_path: str
        def __init__(self, source_path: _Optional[str] = ..., destination_path: _Optional[str] = ...) -> None: ...
    BASE_PATH_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NO_ACL_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    SKIP_CHECKSUM_VALIDATION_FIELD_NUMBER: _ClassVar[int]
    base_path: str
    description: str
    destination_path: str
    encryption_settings: _ydb_export_pb2.EncryptionSettings
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    items: _containers.RepeatedCompositeFieldContainer[ImportFromFsSettings.Item]
    no_acl: bool
    number_of_retries: int
    skip_checksum_validation: bool
    def __init__(self, base_path: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ImportFromFsSettings.Item, _Mapping]]] = ..., description: _Optional[str] = ..., number_of_retries: _Optional[int] = ..., no_acl: bool = ..., skip_checksum_validation: bool = ..., destination_path: _Optional[str] = ..., encryption_settings: _Optional[_Union[_ydb_export_pb2.EncryptionSettings, _Mapping]] = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class ImportFromS3Metadata(_message.Message):
    __slots__ = ["items_progress", "progress", "settings"]
    ITEMS_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    items_progress: _containers.RepeatedCompositeFieldContainer[ImportItemProgress]
    progress: ImportProgress.Progress
    settings: ImportFromS3Settings
    def __init__(self, settings: _Optional[_Union[ImportFromS3Settings, _Mapping]] = ..., progress: _Optional[_Union[ImportProgress.Progress, str]] = ..., items_progress: _Optional[_Iterable[_Union[ImportItemProgress, _Mapping]]] = ...) -> None: ...

class ImportFromS3Request(_message.Message):
    __slots__ = ["operation_params", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    settings: ImportFromS3Settings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ImportFromS3Settings, _Mapping]] = ...) -> None: ...

class ImportFromS3Response(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ImportFromS3Result(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ImportFromS3Settings(_message.Message):
    __slots__ = ["access_key", "bucket", "description", "destination_path", "disable_virtual_addressing", "encryption_settings", "endpoint", "exclude_regexps", "index_filling_mode", "items", "no_acl", "number_of_retries", "region", "scheme", "secret_key", "skip_checksum_validation", "source_prefix"]
    class IndexFillingMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Scheme(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Item(_message.Message):
        __slots__ = ["destination_path", "source_path", "source_prefix"]
        DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PREFIX_FIELD_NUMBER: _ClassVar[int]
        destination_path: str
        source_path: str
        source_prefix: str
        def __init__(self, source_prefix: _Optional[str] = ..., source_path: _Optional[str] = ..., destination_path: _Optional[str] = ...) -> None: ...
    ACCESS_KEY_FIELD_NUMBER: _ClassVar[int]
    BUCKET_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
    DISABLE_VIRTUAL_ADDRESSING_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    HTTP: ImportFromS3Settings.Scheme
    HTTPS: ImportFromS3Settings.Scheme
    INDEX_FILLING_MODE_AUTO: ImportFromS3Settings.IndexFillingMode
    INDEX_FILLING_MODE_BUILD: ImportFromS3Settings.IndexFillingMode
    INDEX_FILLING_MODE_FIELD_NUMBER: _ClassVar[int]
    INDEX_FILLING_MODE_IMPORT: ImportFromS3Settings.IndexFillingMode
    INDEX_FILLING_MODE_UNSPECIFIED: ImportFromS3Settings.IndexFillingMode
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NO_ACL_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    SCHEME_FIELD_NUMBER: _ClassVar[int]
    SECRET_KEY_FIELD_NUMBER: _ClassVar[int]
    SKIP_CHECKSUM_VALIDATION_FIELD_NUMBER: _ClassVar[int]
    SOURCE_PREFIX_FIELD_NUMBER: _ClassVar[int]
    UNSPECIFIED: ImportFromS3Settings.Scheme
    access_key: str
    bucket: str
    description: str
    destination_path: str
    disable_virtual_addressing: bool
    encryption_settings: _ydb_export_pb2.EncryptionSettings
    endpoint: str
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    index_filling_mode: ImportFromS3Settings.IndexFillingMode
    items: _containers.RepeatedCompositeFieldContainer[ImportFromS3Settings.Item]
    no_acl: bool
    number_of_retries: int
    region: str
    scheme: ImportFromS3Settings.Scheme
    secret_key: str
    skip_checksum_validation: bool
    source_prefix: str
    def __init__(self, endpoint: _Optional[str] = ..., scheme: _Optional[_Union[ImportFromS3Settings.Scheme, str]] = ..., bucket: _Optional[str] = ..., access_key: _Optional[str] = ..., secret_key: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ImportFromS3Settings.Item, _Mapping]]] = ..., description: _Optional[str] = ..., number_of_retries: _Optional[int] = ..., region: _Optional[str] = ..., disable_virtual_addressing: bool = ..., no_acl: bool = ..., skip_checksum_validation: bool = ..., source_prefix: _Optional[str] = ..., destination_path: _Optional[str] = ..., encryption_settings: _Optional[_Union[_ydb_export_pb2.EncryptionSettings, _Mapping]] = ..., index_filling_mode: _Optional[_Union[ImportFromS3Settings.IndexFillingMode, str]] = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class ImportItemProgress(_message.Message):
    __slots__ = ["end_time", "parts_completed", "parts_total", "start_time"]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    PARTS_COMPLETED_FIELD_NUMBER: _ClassVar[int]
    PARTS_TOTAL_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    end_time: _timestamp_pb2.Timestamp
    parts_completed: int
    parts_total: int
    start_time: _timestamp_pb2.Timestamp
    def __init__(self, parts_total: _Optional[int] = ..., parts_completed: _Optional[int] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ImportProgress(_message.Message):
    __slots__ = []
    class Progress(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    PROGRESS_BUILD_INDEXES: ImportProgress.Progress
    PROGRESS_CANCELLATION: ImportProgress.Progress
    PROGRESS_CANCELLED: ImportProgress.Progress
    PROGRESS_CREATE_CHANGEFEEDS: ImportProgress.Progress
    PROGRESS_DONE: ImportProgress.Progress
    PROGRESS_PREPARING: ImportProgress.Progress
    PROGRESS_TRANSFER_DATA: ImportProgress.Progress
    PROGRESS_UNSPECIFIED: ImportProgress.Progress
    def __init__(self) -> None: ...

class ListObjectsInFsExportRequest(_message.Message):
    __slots__ = ["operation_params", "page_size", "page_token", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    page_size: int
    page_token: str
    settings: ListObjectsInFsExportSettings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ListObjectsInFsExportSettings, _Mapping]] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ...) -> None: ...

class ListObjectsInFsExportResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListObjectsInFsExportResult(_message.Message):
    __slots__ = ["items", "next_page_token"]
    class Item(_message.Message):
        __slots__ = ["db_path", "fs_path"]
        DB_PATH_FIELD_NUMBER: _ClassVar[int]
        FS_PATH_FIELD_NUMBER: _ClassVar[int]
        db_path: str
        fs_path: str
        def __init__(self, fs_path: _Optional[str] = ..., db_path: _Optional[str] = ...) -> None: ...
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[ListObjectsInFsExportResult.Item]
    next_page_token: str
    def __init__(self, items: _Optional[_Iterable[_Union[ListObjectsInFsExportResult.Item, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...

class ListObjectsInFsExportSettings(_message.Message):
    __slots__ = ["base_path", "encryption_settings", "exclude_regexps", "items", "number_of_retries"]
    class Item(_message.Message):
        __slots__ = ["path"]
        PATH_FIELD_NUMBER: _ClassVar[int]
        path: str
        def __init__(self, path: _Optional[str] = ...) -> None: ...
    BASE_PATH_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    base_path: str
    encryption_settings: _ydb_export_pb2.EncryptionSettings
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    items: _containers.RepeatedCompositeFieldContainer[ListObjectsInFsExportSettings.Item]
    number_of_retries: int
    def __init__(self, base_path: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ListObjectsInFsExportSettings.Item, _Mapping]]] = ..., number_of_retries: _Optional[int] = ..., encryption_settings: _Optional[_Union[_ydb_export_pb2.EncryptionSettings, _Mapping]] = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class ListObjectsInS3ExportRequest(_message.Message):
    __slots__ = ["operation_params", "page_size", "page_token", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    page_size: int
    page_token: str
    settings: ListObjectsInS3ExportSettings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ListObjectsInS3ExportSettings, _Mapping]] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ...) -> None: ...

class ListObjectsInS3ExportResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListObjectsInS3ExportResult(_message.Message):
    __slots__ = ["items", "next_page_token"]
    class Item(_message.Message):
        __slots__ = ["path", "prefix"]
        PATH_FIELD_NUMBER: _ClassVar[int]
        PREFIX_FIELD_NUMBER: _ClassVar[int]
        path: str
        prefix: str
        def __init__(self, prefix: _Optional[str] = ..., path: _Optional[str] = ...) -> None: ...
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[ListObjectsInS3ExportResult.Item]
    next_page_token: str
    def __init__(self, items: _Optional[_Iterable[_Union[ListObjectsInS3ExportResult.Item, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...

class ListObjectsInS3ExportSettings(_message.Message):
    __slots__ = ["access_key", "bucket", "disable_virtual_addressing", "encryption_settings", "endpoint", "exclude_regexps", "items", "number_of_retries", "prefix", "region", "scheme", "secret_key"]
    class Item(_message.Message):
        __slots__ = ["path"]
        PATH_FIELD_NUMBER: _ClassVar[int]
        path: str
        def __init__(self, path: _Optional[str] = ...) -> None: ...
    ACCESS_KEY_FIELD_NUMBER: _ClassVar[int]
    BUCKET_FIELD_NUMBER: _ClassVar[int]
    DISABLE_VIRTUAL_ADDRESSING_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    PREFIX_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    SCHEME_FIELD_NUMBER: _ClassVar[int]
    SECRET_KEY_FIELD_NUMBER: _ClassVar[int]
    access_key: str
    bucket: str
    disable_virtual_addressing: bool
    encryption_settings: _ydb_export_pb2.EncryptionSettings
    endpoint: str
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    items: _containers.RepeatedCompositeFieldContainer[ListObjectsInS3ExportSettings.Item]
    number_of_retries: int
    prefix: str
    region: str
    scheme: ImportFromS3Settings.Scheme
    secret_key: str
    def __init__(self, endpoint: _Optional[str] = ..., scheme: _Optional[_Union[ImportFromS3Settings.Scheme, str]] = ..., bucket: _Optional[str] = ..., access_key: _Optional[str] = ..., secret_key: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ListObjectsInS3ExportSettings.Item, _Mapping]]] = ..., number_of_retries: _Optional[int] = ..., region: _Optional[str] = ..., disable_virtual_addressing: bool = ..., prefix: _Optional[str] = ..., encryption_settings: _Optional[_Union[_ydb_export_pb2.EncryptionSettings, _Mapping]] = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class YdbDumpFormat(_message.Message):
    __slots__ = ["columns"]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    columns: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, columns: _Optional[_Iterable[str]] = ...) -> None: ...



from protos.annotations import sensitive_pb2 as _sensitive_pb2
from protos.annotations import validation_pb2 as _validation_pb2
from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class EncryptionSettings(_message.Message):
    __slots__ = ["encryption_algorithm", "symmetric_key"]
    class SymmetricKey(_message.Message):
        __slots__ = ["key"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        key: bytes
        def __init__(self, key: _Optional[bytes] = ...) -> None: ...
    ENCRYPTION_ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    SYMMETRIC_KEY_FIELD_NUMBER: _ClassVar[int]
    encryption_algorithm: str
    symmetric_key: EncryptionSettings.SymmetricKey
    def __init__(self, encryption_algorithm: _Optional[str] = ..., symmetric_key: _Optional[_Union[EncryptionSettings.SymmetricKey, _Mapping]] = ...) -> None: ...

class ExportItemProgress(_message.Message):
    __slots__ = ["end_time", "parts_completed", "parts_total", "start_time"]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    PARTS_COMPLETED_FIELD_NUMBER: _ClassVar[int]
    PARTS_TOTAL_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    end_time: _timestamp_pb2.Timestamp
    parts_completed: int
    parts_total: int
    start_time: _timestamp_pb2.Timestamp
    def __init__(self, parts_total: _Optional[int] = ..., parts_completed: _Optional[int] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ExportProgress(_message.Message):
    __slots__ = []
    class Progress(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    PROGRESS_CANCELLATION: ExportProgress.Progress
    PROGRESS_CANCELLED: ExportProgress.Progress
    PROGRESS_DONE: ExportProgress.Progress
    PROGRESS_PREPARING: ExportProgress.Progress
    PROGRESS_TRANSFER_DATA: ExportProgress.Progress
    PROGRESS_UNSPECIFIED: ExportProgress.Progress
    def __init__(self) -> None: ...

class ExportToFsMetadata(_message.Message):
    __slots__ = ["items_progress", "progress", "settings"]
    ITEMS_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    items_progress: _containers.RepeatedCompositeFieldContainer[ExportItemProgress]
    progress: ExportProgress.Progress
    settings: ExportToFsSettings
    def __init__(self, settings: _Optional[_Union[ExportToFsSettings, _Mapping]] = ..., progress: _Optional[_Union[ExportProgress.Progress, str]] = ..., items_progress: _Optional[_Iterable[_Union[ExportItemProgress, _Mapping]]] = ...) -> None: ...

class ExportToFsRequest(_message.Message):
    __slots__ = ["operation_params", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    settings: ExportToFsSettings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ExportToFsSettings, _Mapping]] = ...) -> None: ...

class ExportToFsResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ExportToFsResult(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ExportToFsSettings(_message.Message):
    __slots__ = ["base_path", "compression", "description", "encryption_settings", "exclude_regexps", "items", "number_of_retries", "source_path"]
    class Item(_message.Message):
        __slots__ = ["destination_path", "source_path"]
        DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
        destination_path: str
        source_path: str
        def __init__(self, source_path: _Optional[str] = ..., destination_path: _Optional[str] = ...) -> None: ...
    BASE_PATH_FIELD_NUMBER: _ClassVar[int]
    COMPRESSION_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
    base_path: str
    compression: str
    description: str
    encryption_settings: EncryptionSettings
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    items: _containers.RepeatedCompositeFieldContainer[ExportToFsSettings.Item]
    number_of_retries: int
    source_path: str
    def __init__(self, base_path: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ExportToFsSettings.Item, _Mapping]]] = ..., description: _Optional[str] = ..., number_of_retries: _Optional[int] = ..., compression: _Optional[str] = ..., encryption_settings: _Optional[_Union[EncryptionSettings, _Mapping]] = ..., source_path: _Optional[str] = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class ExportToS3Metadata(_message.Message):
    __slots__ = ["items_progress", "progress", "settings"]
    ITEMS_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    items_progress: _containers.RepeatedCompositeFieldContainer[ExportItemProgress]
    progress: ExportProgress.Progress
    settings: ExportToS3Settings
    def __init__(self, settings: _Optional[_Union[ExportToS3Settings, _Mapping]] = ..., progress: _Optional[_Union[ExportProgress.Progress, str]] = ..., items_progress: _Optional[_Iterable[_Union[ExportItemProgress, _Mapping]]] = ...) -> None: ...

class ExportToS3Request(_message.Message):
    __slots__ = ["operation_params", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    settings: ExportToS3Settings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ExportToS3Settings, _Mapping]] = ...) -> None: ...

class ExportToS3Response(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ExportToS3Result(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ExportToS3Settings(_message.Message):
    __slots__ = ["access_key", "bucket", "compression", "description", "destination_prefix", "disable_virtual_addressing", "encryption_settings", "endpoint", "exclude_regexps", "items", "materialize_indexes", "number_of_retries", "region", "scheme", "secret_key", "source_path", "storage_class"]
    class Scheme(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class StorageClass(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class Item(_message.Message):
        __slots__ = ["destination_prefix", "source_path"]
        DESTINATION_PREFIX_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
        destination_prefix: str
        source_path: str
        def __init__(self, source_path: _Optional[str] = ..., destination_prefix: _Optional[str] = ...) -> None: ...
    ACCESS_KEY_FIELD_NUMBER: _ClassVar[int]
    BUCKET_FIELD_NUMBER: _ClassVar[int]
    COMPRESSION_FIELD_NUMBER: _ClassVar[int]
    DEEP_ARCHIVE: ExportToS3Settings.StorageClass
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_PREFIX_FIELD_NUMBER: _ClassVar[int]
    DISABLE_VIRTUAL_ADDRESSING_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTION_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    GLACIER: ExportToS3Settings.StorageClass
    HTTP: ExportToS3Settings.Scheme
    HTTPS: ExportToS3Settings.Scheme
    INTELLIGENT_TIERING: ExportToS3Settings.StorageClass
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    MATERIALIZE_INDEXES_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    ONEZONE_IA: ExportToS3Settings.StorageClass
    OUTPOSTS: ExportToS3Settings.StorageClass
    REDUCED_REDUNDANCY: ExportToS3Settings.StorageClass
    REGION_FIELD_NUMBER: _ClassVar[int]
    SCHEME_FIELD_NUMBER: _ClassVar[int]
    SECRET_KEY_FIELD_NUMBER: _ClassVar[int]
    SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
    STANDARD: ExportToS3Settings.StorageClass
    STANDARD_IA: ExportToS3Settings.StorageClass
    STORAGE_CLASS_FIELD_NUMBER: _ClassVar[int]
    STORAGE_CLASS_UNSPECIFIED: ExportToS3Settings.StorageClass
    UNSPECIFIED: ExportToS3Settings.Scheme
    access_key: str
    bucket: str
    compression: str
    description: str
    destination_prefix: str
    disable_virtual_addressing: bool
    encryption_settings: EncryptionSettings
    endpoint: str
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    items: _containers.RepeatedCompositeFieldContainer[ExportToS3Settings.Item]
    materialize_indexes: bool
    number_of_retries: int
    region: str
    scheme: ExportToS3Settings.Scheme
    secret_key: str
    source_path: str
    storage_class: ExportToS3Settings.StorageClass
    def __init__(self, endpoint: _Optional[str] = ..., scheme: _Optional[_Union[ExportToS3Settings.Scheme, str]] = ..., bucket: _Optional[str] = ..., access_key: _Optional[str] = ..., secret_key: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ExportToS3Settings.Item, _Mapping]]] = ..., description: _Optional[str] = ..., number_of_retries: _Optional[int] = ..., storage_class: _Optional[_Union[ExportToS3Settings.StorageClass, str]] = ..., compression: _Optional[str] = ..., region: _Optional[str] = ..., disable_virtual_addressing: bool = ..., source_path: _Optional[str] = ..., destination_prefix: _Optional[str] = ..., encryption_settings: _Optional[_Union[EncryptionSettings, _Mapping]] = ..., materialize_indexes: bool = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...

class ExportToYtMetadata(_message.Message):
    __slots__ = ["items_progress", "progress", "settings"]
    ITEMS_PROGRESS_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    items_progress: _containers.RepeatedCompositeFieldContainer[ExportItemProgress]
    progress: ExportProgress.Progress
    settings: ExportToYtSettings
    def __init__(self, settings: _Optional[_Union[ExportToYtSettings, _Mapping]] = ..., progress: _Optional[_Union[ExportProgress.Progress, str]] = ..., items_progress: _Optional[_Iterable[_Union[ExportItemProgress, _Mapping]]] = ...) -> None: ...

class ExportToYtRequest(_message.Message):
    __slots__ = ["operation_params", "settings"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    settings: ExportToYtSettings
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., settings: _Optional[_Union[ExportToYtSettings, _Mapping]] = ...) -> None: ...

class ExportToYtResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ExportToYtResult(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ExportToYtSettings(_message.Message):
    __slots__ = ["description", "exclude_regexps", "host", "items", "number_of_retries", "port", "token", "use_type_v3"]
    class Item(_message.Message):
        __slots__ = ["destination_path", "source_path"]
        DESTINATION_PATH_FIELD_NUMBER: _ClassVar[int]
        SOURCE_PATH_FIELD_NUMBER: _ClassVar[int]
        destination_path: str
        source_path: str
        def __init__(self, source_path: _Optional[str] = ..., destination_path: _Optional[str] = ...) -> None: ...
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    EXCLUDE_REGEXPS_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    NUMBER_OF_RETRIES_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    USE_TYPE_V3_FIELD_NUMBER: _ClassVar[int]
    description: str
    exclude_regexps: _containers.RepeatedScalarFieldContainer[str]
    host: str
    items: _containers.RepeatedCompositeFieldContainer[ExportToYtSettings.Item]
    number_of_retries: int
    port: int
    token: str
    use_type_v3: bool
    def __init__(self, host: _Optional[str] = ..., port: _Optional[int] = ..., token: _Optional[str] = ..., items: _Optional[_Iterable[_Union[ExportToYtSettings.Item, _Mapping]]] = ..., description: _Optional[str] = ..., number_of_retries: _Optional[int] = ..., use_type_v3: bool = ..., exclude_regexps: _Optional[_Iterable[str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ydb_cms_v1.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'ydb_cms_v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_cms_pb2 as protos_dot_ydb__cms__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10ydb_cms_v1.proto\x12\nYdb.Cms.V1\x1a\x14protos/ydb_cms.proto2\x9c\x04\n\nCmsService\x12Q\n\x0e\x43reateDatabase\x12\x1e.Ydb.Cms.CreateDatabaseRequest\x1a\x1f.Ydb.Cms.CreateDatabaseResponse\x12Z\n\x11GetDatabaseStatus\x12!.Ydb.Cms.GetDatabaseStatusRequest\x1a\".Ydb.Cms.GetDatabaseStatusResponse\x12N\n\rAlterDatabase\x12\x1d.Ydb.Cms.AlterDatabaseRequest\x1a\x1e.Ydb.Cms.AlterDatabaseResponse\x12N\n\rListDatabases\x12\x1d.Ydb.Cms.ListDatabasesRequest\x1a\x1e.Ydb.Cms.ListDatabasesResponse\x12Q\n\x0eRemoveDatabase\x12\x1e.Ydb.Cms.RemoveDatabaseRequest\x1a\x1f.Ydb.Cms.RemoveDatabaseResponse\x12l\n\x17\x44\x65scribeDatabaseOptions\x12\'.Ydb.Cms.DescribeDatabaseOptionsRequest\x1a(.Ydb.Cms.DescribeDatabaseOptionsResponseBK\n\x15tech.ydb.proto.cms.v1Z2github.com/ydb-platform/ydb-go-genproto/Ydb_Cms_V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_cms_v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\025tech.ydb.proto.cms.v1Z2github.com/ydb-platform/ydb-go-genproto/Ydb_Cms_V1'
  _globals['_CMSSERVICE']._serialized_start=55
  _globals['_CMSSERVICE']._serialized_end=595
# @@protoc_insertion_point(module_scope)



import asyncio
import threading
import time
import typing

import grpc
import pytest

from .common import CallFromSyncToAsync, wrap_set_name_for_asyncio_task
from .._grpc.grpcwrapper.common_utils import (
    GrpcWrapperAsyncIO,
    ServerStatus,
    callback_from_asyncio,
)
from .. import issues

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ydb._grpc.v4.protos import (
        ydb_status_codes_pb2,
        ydb_topic_pb2,
    )
else:
    # noinspection PyUnresolvedReferences
    from ydb._grpc.common.protos import (
        ydb_status_codes_pb2,
        ydb_topic_pb2,
    )


@pytest.fixture()
def separate_loop():
    loop = asyncio.new_event_loop()

    def run_loop():
        loop.run_forever()
        pass

    t = threading.Thread(target=run_loop, name="test separate loop")
    t.start()

    yield loop

    loop.call_soon_threadsafe(lambda: loop.stop())
    t.join()


@pytest.mark.asyncio
class Test:
    async def test_callback_from_asyncio(self):
        class TestError(Exception):
            pass

        def sync_success():
            return 1

        assert await callback_from_asyncio(sync_success) == 1

        def sync_failed():
            raise TestError()

        with pytest.raises(TestError):
            await callback_from_asyncio(sync_failed)

        async def async_success():
            await asyncio.sleep(0)
            return 1

        assert await callback_from_asyncio(async_success) == 1

        async def async_failed():
            await asyncio.sleep(0)
            raise TestError()

        with pytest.raises(TestError):
            await callback_from_asyncio(async_failed)

    async def test_task_name_on_asyncio_task(self):
        task_name = "asyncio task"
        loop = asyncio.get_running_loop()

        async def some_async_task():
            await asyncio.sleep(0)
            return 1

        asyncio_task = loop.create_task(some_async_task())
        wrap_set_name_for_asyncio_task(asyncio_task, task_name=task_name)

        assert asyncio_task.get_name() == task_name


@pytest.mark.asyncio
class TestGrpcWrapperAsyncIO:
    async def test_convert_grpc_errors_to_ydb(self):
        class TestError(grpc.RpcError, grpc.Call):
            def __init__(self):
                pass

            def code(self):
                return grpc.StatusCode.UNAUTHENTICATED

            def details(self):
                return "test error"

        class FromServerMock:
            async def __anext__(self):
                raise TestError()

        wrapper = GrpcWrapperAsyncIO(lambda: None)
        wrapper.from_server_grpc = FromServerMock()

        with pytest.raises(issues.Unauthenticated):
            await wrapper.receive()

    async def convert_status_code_to_ydb_error(self):
        class FromServerMock:
            async def __anext__(self):
                return ydb_topic_pb2.StreamReadMessage.FromServer(
                    status=ydb_status_codes_pb2.StatusIds.OVERLOADED,
                    issues=[],
                )

        wrapper = GrpcWrapperAsyncIO(lambda: None)
        wrapper.from_server_grpc = FromServerMock()

        with pytest.raises(issues.Overloaded):
            await wrapper.receive()


class TestServerStatus:
    def test_success(self):
        status = ServerStatus(
            status=ydb_status_codes_pb2.StatusIds.SUCCESS,
            issues=[],
        )
        assert status.is_success()
        assert issues._process_response(status) is None

    def test_failed(self):
        status = ServerStatus(
            status=ydb_status_codes_pb2.StatusIds.OVERLOADED,
            issues=[],
        )
        assert not status.is_success()
        with pytest.raises(issues.Overloaded):
            issues._process_response(status)


@pytest.mark.asyncio
class TestCallFromSyncToAsync:
    @pytest.fixture()
    def caller(self, separate_loop):
        return CallFromSyncToAsync(separate_loop)

    def test_unsafe_call_with_future(self, separate_loop, caller):
        callback_loop = None

        async def callback():
            nonlocal callback_loop
            callback_loop = asyncio.get_running_loop()
            return 1

        f = caller.unsafe_call_with_future(callback())

        assert f.result() == 1
        assert callback_loop is separate_loop

    def test_unsafe_call_with_result_ok(self, separate_loop, caller):
        callback_loop = None

        async def callback():
            nonlocal callback_loop
            callback_loop = asyncio.get_running_loop()
            return 1

        res = caller.unsafe_call_with_result(callback(), None)

        assert res == 1
        assert callback_loop is separate_loop

    def test_unsafe_call_with_result_timeout(self, separate_loop, caller):
        timeout = 0.01
        callback_loop = None

        async def callback():
            nonlocal callback_loop
            callback_loop = asyncio.get_running_loop()
            await asyncio.sleep(1)
            return 1

        start = time.monotonic()
        with pytest.raises(TimeoutError):
            caller.unsafe_call_with_result(callback(), timeout)
        finished = time.monotonic()

        assert callback_loop is separate_loop
        assert finished - start > timeout

    def test_safe_call_with_result_ok(self, separate_loop, caller):
        callback_loop = None

        async def callback():
            nonlocal callback_loop
            callback_loop = asyncio.get_running_loop()
            return 1

        res = caller.safe_call_with_result(callback(), 1)

        assert res == 1
        assert callback_loop is separate_loop

    def test_safe_call_with_result_timeout(self, separate_loop, caller):
        timeout = 0.01
        callback_loop = None
        cancelled = False

        async def callback():
            nonlocal callback_loop, cancelled
            callback_loop = asyncio.get_running_loop()
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                cancelled = True
                raise

            return 1

        start = time.monotonic()
        with pytest.raises(TimeoutError):
            caller.safe_call_with_result(callback(), timeout)
        finished = time.monotonic()

        # wait one loop for handle task cancelation
        asyncio.run_coroutine_threadsafe(asyncio.sleep(0), separate_loop)

        assert callback_loop is separate_loop
        assert finished - start > timeout
        assert cancelled

    def test_safe_callback_with_0_timeout_ok(self, separate_loop, caller):
        callback_loop = None

        async def f1():
            return 1

        async def f2():
            return await f1()

        async def callback():
            nonlocal callback_loop
            callback_loop = asyncio.get_running_loop()
            return await f2()

        res = caller.safe_call_with_result(callback(), 0)
        assert callback_loop is separate_loop
        assert res == 1

    def test_safe_callback_with_0_timeout_timeout(self, separate_loop, caller):
        callback_loop = None
        cancelled = False

        async def callback():
            try:
                nonlocal callback_loop, cancelled

                callback_loop = asyncio.get_running_loop()
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                cancelled = True
                raise

        with pytest.raises(TimeoutError):
            caller.safe_call_with_result(callback(), 0)

        assert callback_loop is separate_loop
        assert cancelled

    def test_call_sync_ok(self, separate_loop, caller):
        callback_eventloop = None

        def callback():
            nonlocal callback_eventloop
            callback_eventloop = asyncio.get_running_loop()
            return 1

        res = caller.call_sync(callback)
        assert callback_eventloop is separate_loop
        assert res == 1

    def test_call_sync_error(self, separate_loop, caller):
        callback_eventloop = None

        class TestError(RuntimeError):
            pass

        def callback():
            nonlocal callback_eventloop
            callback_eventloop = asyncio.get_running_loop()
            raise TestError

        with pytest.raises(TestError):
            caller.call_sync(callback)
        assert callback_eventloop is separate_loop



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_query_stats.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cprotos/ydb_query_stats.proto\x12\x0eYdb.TableStats\"-\n\x0eOperationStats\x12\x0c\n\x04rows\x18\x01 \x01(\x04\x12\r\n\x05\x62ytes\x18\x02 \x01(\x04\"\xd1\x01\n\x10TableAccessStats\x12\x0c\n\x04name\x18\x01 \x01(\t\x12-\n\x05reads\x18\x03 \x01(\x0b\x32\x1e.Ydb.TableStats.OperationStats\x12/\n\x07updates\x18\x04 \x01(\x0b\x32\x1e.Ydb.TableStats.OperationStats\x12/\n\x07\x64\x65letes\x18\x05 \x01(\x0b\x32\x1e.Ydb.TableStats.OperationStats\x12\x18\n\x10partitions_count\x18\x06 \x01(\x04J\x04\x08\x02\x10\x03\"\xa3\x01\n\x0fQueryPhaseStats\x12\x13\n\x0b\x64uration_us\x18\x01 \x01(\x04\x12\x36\n\x0ctable_access\x18\x02 \x03(\x0b\x32 .Ydb.TableStats.TableAccessStats\x12\x13\n\x0b\x63pu_time_us\x18\x03 \x01(\x04\x12\x17\n\x0f\x61\x66\x66\x65\x63ted_shards\x18\x04 \x01(\x04\x12\x15\n\rliteral_phase\x18\x05 \x01(\x08\"P\n\x10\x43ompilationStats\x12\x12\n\nfrom_cache\x18\x01 \x01(\x08\x12\x13\n\x0b\x64uration_us\x18\x02 \x01(\x04\x12\x13\n\x0b\x63pu_time_us\x18\x03 \x01(\x04\"\xf4\x01\n\nQueryStats\x12\x35\n\x0cquery_phases\x18\x01 \x03(\x0b\x32\x1f.Ydb.TableStats.QueryPhaseStats\x12\x35\n\x0b\x63ompilation\x18\x02 \x01(\x0b\x32 .Ydb.TableStats.CompilationStats\x12\x1b\n\x13process_cpu_time_us\x18\x03 \x01(\x04\x12\x12\n\nquery_plan\x18\x04 \x01(\t\x12\x11\n\tquery_ast\x18\x05 \x01(\t\x12\x19\n\x11total_duration_us\x18\x06 \x01(\x04\x12\x19\n\x11total_cpu_time_us\x18\x07 \x01(\x04\x42R\n\x0etech.ydb.protoZ=github.com/ydb-platform/ydb-go-genproto/protos/Ydb_TableStats\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_query_stats_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\016tech.ydb.protoZ=github.com/ydb-platform/ydb-go-genproto/protos/Ydb_TableStats\370\001\001'
  _OPERATIONSTATS._serialized_start=48
  _OPERATIONSTATS._serialized_end=93
  _TABLEACCESSSTATS._serialized_start=96
  _TABLEACCESSSTATS._serialized_end=305
  _QUERYPHASESTATS._serialized_start=308
  _QUERYPHASESTATS._serialized_end=471
  _COMPILATIONSTATS._serialized_start=473
  _COMPILATIONSTATS._serialized_end=553
  _QUERYSTATS._serialized_start=556
  _QUERYSTATS._serialized_end=800
# @@protoc_insertion_point(module_scope)
